/*
 * Transwarp V1.1, optimized for direction Amiga -> PC with BaudBandit.device
 * Based on Transdisk V4.2
 * Enhancements for the serial port 1998 by Wolfgang Stoeggl
 * Copyright 1995-97 Bernd Schmidt, Marcus Sundberg, Stefan Ropke,
 *                   Rodney Hester, Joanne Dow
 *
 * Use DICE and 2.0 includes or above to compile. SAS C works too.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <strings.h>

#include <exec/devices.h>
#include <exec/io.h>
#include <exec/memory.h>

#include <devices/trackdisk.h>
#include <devices/serial.h>

#include <clib/alib_protos.h>
#include <clib/exec_protos.h>

#include <proto/exec.h>

char *prog_name = "TransWarp";   /* Name of program being run */
char *prog_version = "1.1"; /* Version of the program */
char *serialdevice = "BaudBandit.device";
char *inbuffer;
char *buffer2;

int baudrate;
int readbuffer = 2048;
int error;
int length;
int blength;
int unit = 0;
int serial=0;
int singledrive=0;
int seropen = FALSE;

struct IOStdReq *ioreq;
struct IOExtSer *ior;
struct MsgPort *SerialMP;

static void usage(void);

static void usage(void)
{
    fprintf(stderr, "%s: version %s\n", prog_name, prog_version );
    fprintf(stderr, "Usage: transwarp options\n");
    fprintf(stderr, "Recognized options:\n");
    fprintf(stderr, "-b n: Set the baudrate to n\n");
    fprintf(stderr, "-h: Assume device is high density\n");
    fprintf(stderr, "-1: Single drive mode\n");
    fprintf(stderr, "-d device unit: Use this device instead of DF0:\n");
    fprintf(stderr, "-c device: Use this serial device instead of BaudBandit\n");
    fprintf(stderr, "-r n: Set the serial read buffer to n\n");
    fprintf(stderr, "-s n: Begin transfer at track n\n");
    fprintf(stderr, "-e n: End transfer at track n\n");
    fprintf(stderr, "-w filename: writes the ADF file <filename> to disk\n\n");
    fprintf(stderr, "Example:\n");
    fprintf(stderr, "transwarp >RAM:df1.adf.1 -d trackdisk 1 -s 0 -e 39\n");
    fprintf(stderr, "transfers the first half of the floppy in DF1: into\n");
    fprintf(stderr, "a file in the RAM disk.\n");
    fprintf(stderr, "Or:\n");
    fprintf(stderr, "transwarp -w test.adf\n");
    fprintf(stderr, "writes the ADF-file test.adf to the disk in df0:\n");
}

void    OpenAll( void );
void    CloseAll( void );
void    MotorOff( void );

void
OpenAll()
{
    buffer2 = AllocMem(2*readbuffer, MEMF_CHIP);
    if ((SerialMP = (struct MsgPort *) CreatePort(0,0)) == NULL) {
        printf("Can't create message port.\n");
        CloseAll();
        exit(30);
    }
    if ((ior = (struct IOExtSer *) CreateExtIO(SerialMP, sizeof(struct IOExtSer))) == NULL) {
        printf("Can't allocate storage for ExtIO.\n");
        CloseAll();
        exit(30);
    }
    ior->io_SerFlags |= SERF_SHARED; /* Turn on SHARED Mode */
    if ((error = OpenDevice(serialdevice,0,(struct IORequest*)ior, 0)) != NULL) {
        printf("Can't open %s unit %ld, error=%ld\n", serialdevice, (long) unit, error);
        CloseAll();
        exit(30);
    }
    seropen = TRUE;
    if (baudrate != 0){ /* Set baud rate, parameters if needed. */
        ior->IOSer.io_Command = SDCMD_QUERY;
        ior->IOSer.io_Actual = 0;
        if (DoIO((struct IORequest*)ior) != NULL) {
            printf("Failed SDCMD_QUERY\n");
            CloseAll();
            exit(30);
        }
        if ((length = ior->IOSer.io_Actual) > 0) {
            if ((inbuffer = malloc(length)) == NULL) {
                printf("Out of memory\n");
                CloseAll();
                exit(30);
            }
            ior->IOSer.io_Command = CMD_READ;
            ior->IOSer.io_Length = length;
            ior->IOSer.io_Data = inbuffer;
            if (DoIO((struct IORequest*)ior) != NULL) {
                printf("Failed read.\n");
                free(inbuffer);
                CloseAll();
                exit(30);
            }
           /* fprintf(stderr,"%s\n",inbuffer); /*If you want to see the nag screen */
            free(inbuffer);
        }
        ior->io_Baud = baudrate;
        ior->io_RBufLen = readbuffer;
        ior->IOSer.io_Command = SDCMD_SETPARAMS;
        if (DoIO((struct IORequest*)ior) != NULL) {
            printf("Unable to set parameters.\n");
        }
    }
}

void
CloseAll()
{
    if (seropen)
        CloseDevice((struct IORequest *)ior);
    if (ior)
        DeleteExtIO((struct IORequest *)ior);
    if (SerialMP)
        DeletePort(SerialMP);
    FreeMem(buffer2, 2*readbuffer);
}

void
MotorOff()
{
    ioreq->io_Command = TD_MOTOR;   /* Turn Disk-motor off */
    ioreq->io_Length = 0;
    DoIO( (struct IORequest *) ioreq);
}

int main(int argc, char **argv)
{
    char *filename, *openMode="rb";
    FILE *ADFFile;

    int write=0;
    struct MsgPort *port;
    UBYTE *buffer;
    char devicebuf[256];
    char serdevicebuf[256];
    char *devicename = "trackdisk.device";
    char devicenum = 0;
    int i;
    int starttr = 0, endtr = 79;
    int sectors = 11;

    if (argc == 1) {
        usage();
        exit (1);
    }
    for (i = 1; i < argc;) {
    if (argv[i][0] != '-' || argv[i][2] != 0) {
        usage();
        exit(1);
    }
    switch (argv[i][1]) {
    case 'h':
        sectors = 22;
        i++;
        break;
    case '1':
        singledrive =1;
        i++;
        break;
    case 'd':
        if (i+2 >= argc) {
        usage();
        exit(1);
        }
        devicenum = atoi(argv[i+2]);
        sprintf(devicebuf, "%s.device", argv[i+1]);
        devicename = devicebuf;
        i += 3;
        break;
    case 'c':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        sprintf(serdevicebuf, "%s.device", argv[i+1]);
        serialdevice = serdevicebuf;
        i += 2;
        break;
    case 's':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        starttr = atoi(argv[i+1]);
        i += 2;
        break;
    case 'e':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        endtr = atoi(argv[i+1]);
        i += 2;
        break;
    case 'b':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        baudrate = atoi(argv[i+1]);
            serial =1;
        i += 2;
        break;
    case 'r':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        readbuffer = atoi(argv[i+1]);
        i += 2;
        break;
    case 'w':
        if (i+1 >= argc) {
        usage();
        exit(1);
        }
        filename=argv[i+1];
        write=1;
        i += 2;
        break;
    default:
        usage();
        exit(1);
    }
    }

    if (serial)
      fprintf(stderr,"Using %s, baudrate %d, buffer %d\n", serialdevice, baudrate, readbuffer);
    fprintf(stderr,"Using %s unit %d\n", devicename, devicenum);
    fprintf(stderr,"Tracks are %d sectors\n", sectors);
    fprintf(stderr,"First track %d, last track %d\n", starttr, endtr);

    buffer = AllocMem(512, MEMF_CHIP);
    if (serial)
        OpenAll();
    if (write) {
        if (!serial) {
            ADFFile = fopen(filename,openMode);

            if (!ADFFile) {
                fprintf(stderr,"Error while opening input file\n");
                exit (1);
            }
        }
    }

    if (singledrive) {
        printf("Insert disk and hit RETURN to continue:\n");
        getch();
    }
    port = CreatePort(0, 0);
    if (port) {
    ioreq = CreateStdIO(port);
    if (ioreq) {
        if (OpenDevice(devicename, devicenum, (struct IORequest *) ioreq, 0) == 0) {
        int tr, sec;

        ioreq->io_Command = write ? CMD_WRITE : CMD_READ;
        ioreq->io_Length = 512;
        ioreq->io_Data = buffer;
        for (tr = starttr*2; tr < (endtr+1)*2; tr++) {
            fprintf(stderr,"Track: %d\r",tr/2);
            for (sec = 0; sec < sectors; sec++) {
                fflush(stderr);
                if (write) {
                    if (serial) {               /* begin more complex serial reading */
                        while (blength < 512) {
                            ior->IOSer.io_Command = SDCMD_QUERY;
                            if (DoIO((struct IORequest*)ior) != NULL) {
                                printf("\nFailed SDCMD_QUERY\n");
                                CloseAll();
                                MotorOff();
                                exit(30);
                            }
                            if ((length = ior->IOSer.io_Actual) > 0) {
                                if ((inbuffer = AllocMem(length, MEMF_CHIP)) == NULL) {
                                    printf("\nOut of memory\n");
                                    CloseAll();
                                    MotorOff();
                                    exit(30);
                                }
                                ior->IOSer.io_Command = CMD_READ;
                                ior->IOSer.io_Length = length;
                                ior->IOSer.io_Data = inbuffer;
                                if ((error = DoIO((struct IORequest*)ior)) != NULL) {
                 			        printf("\nFailed read, error=%d\n", error);
                                    FreeMem(inbuffer, length);
                                    CloseAll();
                                    MotorOff();
                                    exit(30);
                                }
                                memmove(buffer2+blength, inbuffer,length);
                                FreeMem(inbuffer, length);
                                blength = blength + length;
                            }
                        }
                        memmove(buffer,buffer2,512);
                        blength = blength-512;
                        memmove(buffer2,buffer2+512,blength);
                    }                           /* end more complex serial reading */
                    if (!serial) {
                        if (fread(buffer, sizeof(UBYTE), 512, ADFFile) < 512) {
                            fprintf(stderr, "\nError: ADF file too short?\n");
                            MotorOff();
                            exit(1);
                        }
                    }
                }
                ioreq->io_Offset = 512 * (tr * sectors + sec);
                DoIO( (struct IORequest *) ioreq);
                if (!write) {
                    if (serial) {
                        ior->IOSer.io_Command = CMD_WRITE;
                        ior->IOSer.io_Length = 512;
                        ior->IOSer.io_Data = buffer;
                        DoIO((struct IORequest *)ior);
                    }
                    if (!serial)
                        fwrite(buffer, sizeof(UBYTE), 512, stdout);
                }
            }
        }
        if (write) {        /* Make sure the last track is written to disk */
            ioreq->io_Command = CMD_UPDATE;
            DoIO( (struct IORequest *) ioreq);
        }
        ioreq->io_Command = TD_MOTOR;   /* Turn Disk-motor off */
        ioreq->io_Length = 0;
        DoIO( (struct IORequest *) ioreq);
        CloseDevice( (struct IORequest *) ioreq);
        } else
        fprintf(stderr,"Unable to open %s unit %d\n", devicename, devicenum);
        DeleteStdIO(ioreq);
    }
    DeletePort(port);
    }
    fprintf(stderr,"\n");
    FreeMem(buffer, 512);
    if (write) {
        if (!serial)
        fclose (ADFFile);
    }
    if (serial)
        CloseAll();
    return 0;
}
